import base64
import binascii
import io
from PIL import Image
import cv2
import numpy as np
from pathlib import Path

def get_base64_from_bytes(bytes_str):
    '''
        Given bytes, returns Base64 representation of those bytes
    '''
    return base64.b64encode(bytes_str).decode()

def get_jpeg_bytes_from_rgb_img(numpy_image):
    '''
        Given a numpy image, compresses to the image to JPEG and returns the bytes
    '''
    buf = io.BytesIO()
    pillow_image = Image.fromarray(numpy_image)
    pillow_image.save(buf, format='JPEG')
    byte_im = buf.getvalue()
    return byte_im

def get_jpeg_bytes_from_pillow_img(pillow_image):
    '''
        Given a numpy image, compresses to the image to JPEG and returns the bytes
    '''
    buf = io.BytesIO()
    pillow_image.save(buf, format='JPEG')
    byte_im = buf.getvalue()
    return byte_im

def get_jpeg_bytes_from_bgr_img(numpy_image):
    '''
        Given a numpy image, compresses to the image to JPEG and returns the bytes
    '''
    rgb_image = cv2.cvtColor(numpy_image, cv2.COLOR_BGR2RGB)
    return get_jpeg_bytes_from_rgb_img(rgb_image)

def get_pillow_image_from_base64(base64_string):
    '''
        Given a Base64 string for an image, returns a numpy array in RGB order
    '''
    nparr = np.frombuffer(base64.b64decode(str(base64_string)), np.uint8)
    result = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    if result is None:
        raise binascii.Error()
    result = cv2.cvtColor(result, cv2.COLOR_BGR2RGB)
    pillow_image = Image.fromarray(result)
    return pillow_image

def normalize_image(image):
    '''
    Given a single image that could be numpy, pillow, path string or path,
    returns the Pillow image
    '''
    if isinstance(image, np.ndarray):
        try:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            pillow_image = Image.fromarray(image)
        except:
            raise ValueError("image is not in BGR format")
    elif isinstance(image, Image.Image):
        return image
    elif isinstance(image, Path):
        try:
            path = str(image)
            n_image = cv2.imread(path)
            n_image = cv2.cvtColor(n_image, cv2.COLOR_BGR2RGB)
            if n_image is None:
                raise ValueError("Could not read image from path: {}".format(path))
            pillow_image = Image.fromarray(n_image)
        except:
            raise ValueError("Could not read image from path: {}".format(path))
    elif isinstance(image, str):
        try:
            path = str(image)
            n_image = cv2.imread(path)
            n_image = cv2.cvtColor(n_image, cv2.COLOR_BGR2RGB)
            if n_image is None:
                raise ValueError("Could not read image from path: {}".format(path))
            pillow_image = Image.fromarray(n_image)
        except:
            raise ValueError("Could not read image from path: {}".format(path))
    else:
        raise TypeError("At least one image is of inappropriate type")
    return pillow_image


def normalize_images(images):
    '''
    Given an array of images that could be numpy, pillow, path string or path,
    returns an array of Pillow images
    '''
    result = []
    for image in images:
        pillow_image = normalize_image(image)
        result.append(pillow_image)

    return result


import requests
from .api_error import APIError, APIDataValidationError
from typing import List, Union

class HttpInteractor:
    """Class to help with HTTP interactions
    :param backend_url: the URL where the REST API is deployed
    :type str
    :param developer_key: the developer key to use while interacting with the API
    :type str
    """
    def __init__(self, backend_url:str, developer_key:str) -> None:
        self.__backend_url = backend_url
        self.__developer_key = developer_key

    def get(self, url_fragment: str, params:dict=None) -> Union[dict, List[dict]]:
        """Performs an HTTP GET request

        :param url_fragment: the URL fragment from the base URL
        :type url_fragment: str
        :param params: if specified, the parameters to pass to the call, defaults to None
        :type params: Optional[dict]

        :raises APIError: an API7 error with http status code, error code, and error message

        :return: a dictionary or a list of dictionaries
        :rtype: Union[dict, List[dict]]
        """
        url = f"{self.__backend_url}/{url_fragment}"
        headers = {"x-api-key": self.__developer_key}
        response = None
        if params:
            response = requests.get(url, params=params, headers=headers)
        else:
            response = requests.get(url, headers=headers)

        if response.status_code == 422:
            body = response.json()
            error = APIDataValidationError(body)
            raise error
        elif response.status_code != 200:
            body = response.json()
            error = APIError(response.status_code, body["code"], body["message"])
            if "retry-after" in response.headers:
                error.retry_after = int(response.headers["retry-after"])
            raise error

        return response.json()

    def post(self, url_fragment: str, obj:dict, expected_status=201) -> dict:
        """Performs an HTTP POST request

        :param url_fragment: the URL fragment from the base URL
        :type url_fragment: str
        :param obj: the object to be posted represented as a dictionary
        :type obj: dict
        :param expected_status: the expected status code, defaults to 201
        :type expected_status: int

        :raises APIError: an API7 error with http status code, error code, and error message

        :return: a dictionary
        :rtype: dict
        """
        url = f"{self.__backend_url}/{url_fragment}"
        headers = {"x-api-key": self.__developer_key}
        response = requests.post(url, json=obj, headers=headers)

        if response.status_code == 422:
            body = response.json()
            error = APIDataValidationError(body)
            raise error
        elif response.status_code != expected_status:
            body = response.json()
            error = APIError(response.status_code, body["code"], body["message"])
            if "retry-after" in response.headers:
                error.retry_after = int(response.headers["retry-after"])
            raise error

        return response.json()

    def patch(self, url_fragment: str, obj:dict) -> dict:
        """Performs an HTTP PATCH request

        :param url_fragment: the URL fragment from the base URL
        :type url_fragment: str
        :param obj: the object to be posted represented as a dictionary
        :type obj: dict

        :raises APIError: an API7 error with http status code, error code, and error message

        :return: a dictionary
        :rtype: dict
        """
        url = f"{self.__backend_url}/{url_fragment}"
        headers = {"x-api-key": self.__developer_key}
        response = requests.patch(url, json=obj, headers=headers)

        if response.status_code == 422:
            body = response.json()
            error = APIDataValidationError(body)
            raise error
        elif response.status_code != 200:
            body = response.json()
            error = APIError(response.status_code, body["code"], body["message"])
            if "retry-after" in response.headers:
                error.retry_after = int(response.headers["retry-after"])
            raise error

        return response.json()

    def delete(self, url_fragment: str, id:str) -> None:
        """Performs an HTTP DELETE request

        :param url_fragment: the URL fragment from the base URL
        :type url_fragment: str
        :param id: the id of the object to be deleted
        :type id: str

        :raises APIError: an API7 error with http status code, error code, and error message
        """
        url = f"{self.__backend_url}/{url_fragment}/{id}"
        headers = {"x-api-key": self.__developer_key}
        response = requests.delete(url, headers=headers)

        if response.status_code == 422:
            body = response.json()
            error = APIDataValidationError(body)
            raise error
        elif response.status_code != 200:
            body = response.json()
            error = APIError(response.status_code, body["code"], body["message"])
            if "retry-after" in response.headers:
                error.retry_after = int(response.headers["retry-after"])
            raise error





from typing import Union
class APIError(Exception):
    """Class that defines an API Error
    :param http_status_code: the HTTP status code of the error
    :type int
    :param err_code: the API7 error code
    :type str
    :param messsage: the API7 error message
    :type str
    :param retry_after: if present, specifies the time in seconds when the request can be retried to meet rate limit checks
    :type Union[int, None]
    """
    def __init__(self, http_status_code:int, err_code:str, messsage: str, retry_after:Union[int, None]=None) -> None:
        self.http_status_code = http_status_code
        self.err_code = err_code
        self.messsage = messsage
        self.retry_after = retry_after

class APIDataValidationError(Exception):
    """Class that defines an API Data Validation Error
    :param details: Error details from API Server
    :type dict
    """
    def __init__(self, details) -> None:
        self.details = details


from .collections.collections_manager import CollectionsManager
from .persons.persons_manager import PersonsManager
from .search.search_manager import SearchManager
from .liveness.liveness_manager import LivenessManager
from .compare.compare_manager import CompareManager
from .http_interactor import HttpInteractor

class SDK:
    """Class to interact with the API7 REST API
    :param backend_url: the URL where the REST API is deployed
    :type str
    :param developer_key: the developer key to use while interacting with the API
    :type str
    """
    def __init__(self, backend_url:str, developer_key: str) -> None:
        interactor = HttpInteractor(backend_url, developer_key)
        self.collections = CollectionsManager(interactor)
        self.persons = PersonsManager(interactor)
        self.search = SearchManager(interactor)
        self.liveness = LivenessManager(interactor)
        self.compare = CompareManager(interactor)
    


    


import copy
from typing import List, Union, Optional
import numpy as np
import cv2
from PIL.Image import Image
from pathlib import Path
import enum
from datetime import datetime
from ..collections.schemas import Collection, CollectionBase
from .. import util as util 
from dateutil.parser import parse

'''
Gender of the person
'''
class PersonGender(enum.Enum):
    M = "M"
    F = "F"

'''
Gender of the person
'''
class PersonOrderBy(enum.Enum):
    NAME = "name"
    DATE_OF_BIRTH = "date_of_birth"
    GENDER = "gender"
    NATIONALITY = "nationality"
    MODIFIED_DATE = "modified_date"
    CREATE_DATE = "create_date"


class PersonBase:
    '''A person object

    :param images: a list of images (max 3), with each image being either a numpy array (obtained with cv2.imread), a string (path to a file), a Pillow image (obtained with PIL.Image.open()), or a pathlib path
    :type images:List[Union[np.ndarray, str, Image, Path]]
    :param id: a unique identifier for the person. If unspecified a new id will be generated, defaults to None
    :type id: Optional[str]
    :param name: the name of the person
    :type name:Optional[str]
    :param gender: the gender of the person, defaults to None
    :type gender: Optional[PersonGender]
    :param date_of_birth: the date of birth of the person, defaults to None
    :type date_of_birth: Optional[datetime]
    :param nationality: the nationality of the person, defaults to None
    :type nationality: Optional[str]
    :param collections: a list containing either Collection, CollectionBase objects, or strings representing the id of the collections that this person belongs to, defaults to empty list []
    :type collections: Optional[List[Union[Collection, CollectionBase, str]]]
    :param notes: any additional notes about the person, defaults to None
    :type notes: Optional[str]
    :param is_bulk_insert: whether this person is part of a bulk insert batch, defaults to False
    :type is_bulk_insert: Optional[bool]
    '''
    def __init__(self, 
                 images: List[Union[np.ndarray, str, Image, Path]],
                 id:Optional[str] = None,
                 name: Optional[str] = None, 
                 gender:Optional[PersonGender] = None,
                 date_of_birth:Optional[datetime] = None,
                 nationality:Optional[str] = None,
                 collections:Optional[List[Union[Collection, CollectionBase, str]]]=[],
                 notes:Optional[str]=None,
                 is_bulk_insert:Optional[bool] = False) -> None:
        if not isinstance(images, list):
            raise TypeError("images must be a list")
        if len(images) > 3:
            raise ValueError("images list cannot contain more than 3 images")
        if len(images) < 1:
            raise ValueError("images list cannot be empty")
        self.name = name
        self._images = util.normalize_images(images)
        self.id = id
        self.gender = gender
        self.date_of_birth = date_of_birth
        self.nationality = nationality
        self.collections = collections
        self.notes = notes
        self.is_bulk_insert = is_bulk_insert
        self.__server_object:Person = None

    @property
    def images(self):
        raise RuntimeError('This property has no getter. Please use get_image(index:int) -> Image')

    @images.setter
    def images(self, images:List[Union[np.ndarray, str, Image, Path]]):
        """Replaces the current set of images with the ones specified

        :param images: the list of images to be set for this person, each image must be at least 224x224 pixels
        :type images:List[Union[np.ndarray, str, Image, Path]]
        """
        if not isinstance(images, list):
            raise TypeError("images must be a list")
        if len(images) > 3:
            raise ValueError("images list cannot contain more than 3 images")
        if len(images) < 1:
            raise ValueError("images list cannot be empty")
        images:List[Image] = util.normalize_images(images)
        for image in images:
            if image.height < 224 or image.width < 224:
                raise ValueError("At least one image is smaller than 224x224")
        self._images = images

    def get_image(self, image_index: int) -> Image:
        """Get's the image at the specified index

        :param image_index: the index of the image to get
        :type image_index:int
        """
        return self._images[image_index]

    @property
    def name(self):
        '''Name of the person'''
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, str) and len(value) == 0:
            raise ValueError("If specified, name must be a non-empty string")
        elif value is not None and not isinstance(value, str):
            raise TypeError("If specified, name must be a string")
        self._name = value

    @property
    def id(self):
        '''Id of the person'''
        return self._id

    @id.setter
    def id(self, value):
        if isinstance(value, str) and len(value) == 0:
            raise ValueError("id must be a non-empty string or None")
        if not isinstance(value, str) and not value is None:
            raise TypeError("id must be a string or None")
        self._id = value

    @property
    def gender(self):
        '''Gender of the person'''
        return self._gender

    @gender.setter
    def gender(self, value):
        if value is not None and not isinstance(value, PersonGender):
            raise TypeError("gender must be of type PersonGender or None")
        self._gender = value

    @property
    def date_of_birth(self):
        '''Date of birth of the person'''
        return self._date_of_birth

    @date_of_birth.setter
    def date_of_birth(self, value):
        if value is not None and not isinstance(value, datetime):
            raise TypeError("date_of_birth must be of type datetime or None")
        self._date_of_birth = value

    @property
    def nationality(self):
        '''Nationality of the person'''
        return self._nationality

    @nationality.setter
    def nationality(self, value):
        if not isinstance(value, str) and not value is None:
            raise TypeError("nationality must be of type str or None")
        self._nationality = value

    @property
    def collections(self):
        '''Collections that the person belongs to'''
        return self._collections

    @collections.setter
    def collections(self, value):
        if not isinstance(value, list):
            raise TypeError("collections must be of type list")
        for collection in value:
            if not isinstance(collection, Collection) and not isinstance(collection, CollectionBase):
                raise TypeError("At least one item in collections is neither of type Collection nor CollectionBase")
        self._collections = value

    @property
    def notes(self):
        '''Any notes related to the person'''
        return self._notes

    @notes.setter
    def notes(self, value):
        if not isinstance(value, str) and not value is None:
            raise TypeError("notes must be of type str or None")
        self._notes = value

    @property
    def is_bulk_insert(self):
        '''Whether this person is part of a bulk insert operation'''
        return self._is_bulk_insert

    @is_bulk_insert.setter
    def is_bulk_insert(self, value):
        if not isinstance(value, bool):
            raise TypeError("is_bulk_insert must be of type bool")
        self._is_bulk_insert = value

    def set_server_obj(self, obj):
        self.__server_object = obj

    def to_dict(self) -> dict:
        '''
        Return a dictionary representation of this object

        :return: a dictionary representing this object
        :rtype: dict
        '''
        repr = {}

        if not self.__server_object:
            # If there is no server object, images are always included
            images = util.normalize_images(self._images)
            repr["images"] = images
        else:
            # Otherwise, include only if they are changed
            if self.name != self.__server_object.name:
                repr["name"] = self.name

            if self._images != self.__server_object._images:
                images = util.normalize_images(self._images)
                repr["images"] = images

        # Id should always be included
        repr["id"] = self.id

        if not self.__server_object:
            if self.gender == PersonGender.M:
                repr["gender"] = "M"
            elif self.gender == PersonGender.F:
                repr["gender"] = "F"
            else:
                repr["gender"] = None
        else:
            if self.__server_object.gender != self.gender:
                if self.gender == PersonGender.M:
                    repr["gender"] = "M"
                elif self.gender == PersonGender.F:
                    repr["gender"] = "F"
                else:
                    repr["gender"] = None
        

        if not self.__server_object:
            if self.date_of_birth is not None:
                repr["date_of_birth"] = str(self.date_of_birth).split(' ')[0]
            else:
                repr["date_of_birth"] = None
        else:
            if self.__server_object.date_of_birth != self.date_of_birth:
                if self.date_of_birth is not None:
                    repr["date_of_birth"] = str(self.date_of_birth).split(' ')[0]
                else:
                    repr["date_of_birth"] = None

        if not self.__server_object:
            repr["nationality"] = self.nationality
        else:
            if self.__server_object.nationality != self.nationality:
                repr["nationality"] = self.nationality

        if not self.__server_object:
            repr["name"] = self.name
        else:
            if self.__server_object.name != self.name:
                repr["name"] = self.name

        if not self.__server_object:
            collections = []
            for collection in self.collections:
                if isinstance(collection, CollectionBase):
                    collections.append(collection.id)
                elif isinstance(collection, Collection):
                    collections.append(collection.id)
                elif isinstance(collection, str):
                    collections.append(collection)
                else:
                    raise TypeError("At least one collection is of inappropriate type")
            
            repr["collections"] = collections
        else:
            if self.__server_object.collections != self.collections:
                collections = []
                for collection in self.collections:
                    if isinstance(collection, CollectionBase):
                        collections.append(collection.id)
                    elif isinstance(collection, Collection):
                        collections.append(collection.id)
                    elif isinstance(collection, str):
                        collections.append(collection)
                    else:
                        raise TypeError("At least one collection is of inappropriate type")
                repr["collections"] = collections


        if not self.__server_object:
            repr["notes"] = self.notes
        else:
            if self.__server_object.notes != self.notes:
                repr["notes"] = self.notes

        if not self.__server_object:
            repr["is_bulk_insert"] = self.is_bulk_insert
        else:
            if self.is_bulk_insert != self.is_bulk_insert:
                repr["is_bulk_insert"] = self.is_bulk_insert

        return repr

    # def __getattribute__(self, name):
    #     val = object.__getattribute__(self, name)
    #     return val

    def __repr__(self) -> str:
        return str({
            "id": self.id,
            "name": self.name,
            "images": self._images,
            "gender": self.gender,
            "collections": self.collections,
            "date_of_birth": self.date_of_birth,
            "nationality": self.nationality,
            "notes": self.notes
        })

class Person(PersonBase):
    '''A previously created person object

    :param images: a list of images (max 3), with each image being either a numpy array (obtained with cv2.imread), a string (path to a file), a Pillow image (obtained with PIL.Image.open()), or a pathlib path
    :type images:List[Union[numpy.ndarray, str, PIL.Image.Image, pathlib.Path]]
    :param id: a unique identifier for the person. If unspecified a new id will be generated, defaults to None
    :type id: Optional[str]
    :param name: the name of the person
    :type name:Optional[str]
    :param gender: the gender of the person, defaults to None
    :type gender: Optional[PersonGender]
    :param date_of_birth: the date of birth of the person, defaults to None
    :type date_of_birth: Optional[datetime]
    :param nationality: the nationality of the person, defaults to None
    :type nationality: Optional[str]
    :param collections: a list containing either Collection, CollectionBase objects, or strings representing the id of the collections that this person belongs to, defaults to empty list []
    :type collections: Optional[List[Union[Collection, CollectionBase, str]]]
    :param notes: any additional notes about the person, defaults to None
    :type notes: Optional[str]
    :param is_bulk_insert: whether this person is part of a bulk insert batch, defaults to False
    :type is_bulk_insert: Optional[bool]
    :param create_date: date that this person was created
    :type create_date: datetime
    :param modified_date: date that this person was modified
    :type modified_date: datetime
    '''
    def __init__(self, 
                 images: List[Union[np.ndarray, str, Image, Path]],
                 id:Optional[str] = None,
                 name: Optional[str] = None, 
                 gender:Optional[PersonGender] = None,
                 date_of_birth:Optional[datetime] = None,
                 nationality:Optional[str] = None,
                 collections:Optional[List[Union[Collection, CollectionBase, str]]]=[],
                 notes:Optional[str]=None,
                 is_bulk_insert:Optional[bool] = False,
                 create_date:datetime = None,
                 modified_date:datetime = None) -> None:
        super().__init__(images, id=id, name=name, gender=gender,
                        date_of_birth=date_of_birth, nationality=nationality,
                        collections=collections, notes=notes,
                        is_bulk_insert=is_bulk_insert)

        if not isinstance(create_date, datetime):
            raise TypeError("create_date must be of type datetime")

        if not isinstance(modified_date, datetime):
            raise TypeError("modified_date must be of type datetime")

        self._create_date = create_date
        self._modified_date = modified_date

    @property
    def create_date(self):
        '''Date that the person was created'''
        return self._create_date

    @property
    def modified_date(self):
        '''Date that the person was modified'''
        return self._modified_date

    @classmethod
    def from_dict(cls, obj:dict):
        """Get a person initialized from a dict

        :param obj: the dictionary representing this object
        :type obj:dict

        :return: a Person object
        :rtype: Person
        """
        collections = [Collection.from_dict(item) for item in obj["collections"]]
        thumbnails = [util.get_pillow_image_from_base64(item["thumbnail"]) for item in obj["thumbnails"]]
        gender = obj["gender"]
        if gender == "M":
            gender = PersonGender.M
        elif gender == "F":
            gender = PersonGender.F
        else:
            gender = None
        server_obj = Person( 
            thumbnails,
            id=obj["id"],
            name=obj["name"],
            gender=gender,
            date_of_birth= parse(obj["date_of_birth"]) if obj["date_of_birth"] else None,
            nationality=obj["nationality"],
            collections=collections,
            notes=obj["notes"],
            is_bulk_insert=False,
            create_date=parse(obj["create_date"]),
            modified_date=parse(obj["modified_date"]))

        client_obj = copy.copy(server_obj)
        client_obj._images = thumbnails
        client_obj.set_server_obj(server_obj)
        return client_obj

    def __repr__(self) -> str:
        return str({
            "id": self.id,
            "name": self.name,
            "images": self._images,
            "gender": self.gender,
            "collections": self.collections,
            "date_of_birth": self.date_of_birth,
            "nationality": self.nationality,
            "notes": self.notes,
            "create_date": str(self.create_date),
            "modified_date": str(self.modified_date)
        })

class PersonList:
    '''A list of persons
    
    :param count: the total count of all persons in the database
    :type count:int
    :param persons: the persons matching the criteria
    :type persons:List[Person]
    '''
    
    def __init__(self, count: int, persons: List[Person]) -> None:
        if not isinstance(count, int):
            raise TypeError("count must be of type int")
        if not isinstance(persons, list):
            raise TypeError("persons must be of type list")
        for person in persons:
            if not isinstance(person, Person):
                raise TypeError("At least one item in persons is not an object of type Person")
        self._count: int = count
        self._persons: List[Person] = persons

    @property
    def count(self):
        '''Total number of persons in the database that match the criteria'''
        return self._count

    @property
    def persons(self):
        '''The (paged) list of persons that match the criteria'''
        return self._persons

    @classmethod
    def from_dict(cls, obj:dict):
        """Get a person list initialized from a dict

        :param obj: the dictionary representing this object
        :type obj:dict

        :return: a PersonList object
        :rtype: PersonList
        """
        persons = [Person.from_dict(item) for item in obj["persons"]]
        return PersonList(obj["count"], persons)

    def to_dict(self) -> dict:
        '''
        Return a dictionary representation of this object

        :return: a dictionary representing this object
        :rtype: dict
        '''
        return {
            "count": self.count,
            "persons": [item.to_dict() for item in self.persons],
        }

    def __repr__(self) -> str:
        return str({"count": self.count, "persons": [str(item) for item in self.persons]})

        

    

    



from ..http_interactor import HttpInteractor
from ..persons.schemas import Person, PersonList, PersonOrderBy
from .. import util as util
from ..schemas import SortOrder
from typing import Optional

class PersonsManager:
    '''Class to manage persons
    
    :param http_interactor: the http interactor instance
    :type http_interactor:HttpInteractor
    '''
    
    def __init__(self, http_interactor:HttpInteractor) -> None:
        self.__http_interactor = http_interactor

    def list(self, skip:Optional[int]=0, 
             take:Optional[int]=20, 
             order_by:Optional[PersonOrderBy]=PersonOrderBy.NAME, 
             order:Optional[SortOrder]=SortOrder.ASC, 
             search:Optional[str]=None) -> PersonList:
        """Get a persons list 

        :param skip: skip the specified number of initial rows, defaults to 0
        :type skip:Optional[int]
        :param take: get the specified number of rows, defaults to 20
        :type take:Optional[int]
        :param order: order ascending(SortOrder.ASC) or descending(SortOrder.DESC)
        :type order:Optional[SortOrder]
        :param order_by: the column to choose for the ordering
        :type order_by:Optional[PersonOrderBy]
        :param search: a search string to filter collections by name/description
        :type search:Optional[str]

        :raises ApiError: an API error if unsuccessful

        :return: a CollectionList object
        :rtype: CollectionList
        """
        s_order_by = str(order_by).lower().replace("personorderby.", "")
        order = "ASC" if order == SortOrder.ASC else "DESC"
        params = {"skip": skip, "take": take, "order": order, "search": search, "order_by": s_order_by}
        result = self.__http_interactor.get("persons", params)
        return PersonList.from_dict(result)

    def create(self, person:Person) -> Person:
        """Create a person

        :param person: a person object minimally specifying a name and images
        :type person:Person

        :raises ApiError: an API error if unsuccessful

        :return: a Person object representing the person that was created
        :rtype: Person
        """
        post_body = person.to_dict()
        b64_images = []
        for image in post_body["images"]:
            i_bytes = util.get_jpeg_bytes_from_pillow_img(image)
            i_str = util.get_base64_from_bytes(i_bytes)
            b64_images.append(i_str)
        post_body["images"] = b64_images

        result = self.__http_interactor.post("person", post_body)
        return Person.from_dict(result)

    def update(self, person:Person) -> Person:
        """Updates a person

        :param person: a person object
        :type person:Person

        :raises ApiError: an API error if unsuccessful

        :return: a Person object representing the person that was updated
        :rtype: Person
        """
        obj = person.to_dict()
        if len(obj.keys()) == 1:
            raise ValueError("Object is not modified. Nothing to update")
        b64_images = []
        if "images" in obj:
            for image in obj["images"]:
                i_bytes = util.get_jpeg_bytes_from_pillow_img(image)
                i_str = util.get_base64_from_bytes(i_bytes)
                b64_images.append(i_str)
            obj["images"] = b64_images
        result = self.__http_interactor.patch("person", obj)
        result = Person.from_dict(result)
        return result

    def delete(self, id:str) -> None:
        """Deletes the specified person

        :param id: the id of the person to be deleted
        :type id:str

        :raises ApiError: an API error if unsuccessful
        """
        self.__http_interactor.delete("person", id)

    def get(self, id:str) -> Person:
        """Gets person specified by the id

        :param id: the id of the person to be retrieved
        :type id:str

        :raises ApiError: an API error if unsuccessful

        :return: a Person object
        :rtype: Person
        """
        result = self.__http_interactor.get(f"person/{id}")
        return Person.from_dict(result)

    



from ..http_interactor import HttpInteractor
from ..schemas import SortOrder
from ..collections.schemas import CollectionBase, CollectionList, Collection
from typing import Optional


class CollectionsManager:
    """Class to help manage collections
    
    :param http_interactor: the http interactor instance
    :type http_interactor:HttpInteractor
    """
    def __init__(self, http_interactor:HttpInteractor) -> None:
        self.__http_interactor = http_interactor

    def list(self, skip:Optional[int]=0, 
             take:Optional[int]=20, 
             order:Optional[SortOrder]=SortOrder.ASC, 
             search:Optional[str]=None) -> CollectionList:
        """Get a collections list 

        :param skip: skip the specified number of initial rows, defaults to 0
        :type skip:Optional[int]
        :param take: get the specified number of rows, defaults to 20
        :type take:Optional[int]
        :param order: order ascending(SortOrder.ASC) or descending(SortOrder.DESC)
        :type order:Optional[SortOrder]
        :param search: a search string to filter collections by name/description
        :type search:Optional[str]

        :raises API7.ApiError: an API error if unsuccessful

        :return: a CollectionList object
        :rtype: CollectionList
        """
        if not isinstance(order, SortOrder):
            raise TypeError("order should be of type SortOrder Enum")

        order = "ASC" if order == SortOrder.ASC else "DESC"
        params = {"skip": skip, "take": take, "order": order, "search": search}
        result = self.__http_interactor.get("collections", params)
        return CollectionList.from_dict(result)

    def create(self, collection:CollectionBase) -> Collection:
        """Create a collection

        :param collection: a collection base object specifying the name and description of the collection to be created
        :type collection:CollectionBase

        :raises API7.ApiError: an API error if unsuccessful

        :return: a Collection object representing the collection that was created
        :rtype: Collection
        """
        obj = collection.to_dict()
        if "count" in obj:
            del obj["count"]
        result = self.__http_interactor.post("collection", obj)
        result = Collection.from_dict(result)
        return result

    def update(self, collection:Collection) -> Collection:
        """Updates a collection

        :param collection: a collection object specifying the id, name and description of the collection to be updated
        :type collection:Collection

        :raises ApiError: an API error if unsuccessful

        :return: a Collection object representing the collection that was updated
        :rtype: Collection
        """
        obj = collection.to_dict()
        if "count" in obj:
            del obj["count"]
        result = self.__http_interactor.patch("collection", obj)
        result = Collection.from_dict(result)
        return result

    def delete(self, id:str) -> None:
        """Deletes the specified collection

        :param id: the id of the collection to be deleted
        :type id:str

        :raises ApiError: an API error if unsuccessful
        """
        self.__http_interactor.delete("collection", id)

    def get(self, id:str) -> Collection:
        """Gets collection specified by the id

        :param id: the id of the collection to be retrieved
        :type id:str

        :raises ApiError: an API error if unsuccessful

        :return: a Collection object
        :rtype: Collection
        """
        result = self.__http_interactor.get(f"collection/{id}")
        return Collection.from_dict(result)


from datetime import datetime
from typing import List
from dateutil.parser import parse

class CollectionBase:
    '''A minimal collection object for creation

    :param name: the name of the collection
    :type name:str
    :param description: the description of the collection
    :type description: str
    '''
    def __init__(self, name: str, description: str = None) -> None:
        self.name = name
        self._description: str = description

    @property
    def name(self):
        '''Name of the collection'''
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("name accepts only str")
        if not isinstance(value, str) or len(value) == 0:
            raise ValueError("name must be a non-empty string")
        self._name = value

    @property
    def description(self):
        '''Description of the collection'''
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("description must be a string or None")
        self._description = value

    def to_dict(self) -> dict:
        '''
        Return a dictionary representation of this object

        :return: a dictionary representing this object
        :rtype: dict
        '''
        return {
            "name": self.name,
            "description": self.description
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

class Collection(CollectionBase):
    '''A previously created collection object

    :param id: the id of a previously created collection
    :type id:str
    :param name: the name of the collection
    :type name:str
    :param description: the description of the collection
    :type description: str
    :param count: the number of persons in this collection
    :type count: int
    :param create_date: date that this collection was created
    :type create_date: datetime
    :param modified_date: date that this collection was modified
    :type modified_date: datetime
    '''
    def __init__(self, id: str, name: str, description: str, count: int,
                 create_date: datetime, modified_date: datetime) -> None:
        super().__init__(name, description)

        if not isinstance(id, str):
            raise TypeError("id must be non-empty string")
        if len(id) == 0:
            raise ValueError("id must be a non-empty string")

        if not isinstance(count, int):
            raise TypeError("count must be an int")

        if not isinstance(create_date, datetime):
            raise TypeError("create_date must be a datetime")

        if not isinstance(modified_date, datetime):
            raise TypeError("modified_date must be a datetime")

        self._id: str = id
        self._count = count
        self._create_date = create_date
        self._modified_date = modified_date

    @property
    def count(self):
        '''The number of persons in this collection'''
        return self._count

    @property
    def id(self):
        '''The id of this collection'''
        return self._id

    @property
    def create_date(self):
        '''The date this collection was created'''
        return self._create_date

    @property
    def modified_date(self):
        '''The date this collection was modified'''
        return self._modified_date

    @classmethod
    def from_dict(cls, obj:dict):
        """Get a collection initialized from a dict

        :param obj: the dictionary representing this object
        :type obj:dict

        :return: a Collection object
        :rtype: Collection
        """
        return Collection(obj["id"], obj["name"], obj["description"], obj["count"],
            parse(obj["create_date"]), parse(obj["modified_date"]))

    def to_dict(self) -> dict:
        '''
        Return a dictionary representation of this object

        :return: a dictionary representing this object
        :rtype: dict
        '''
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "count": self.count,
            "create_date": str(self.create_date),
            "modified_date": str(self.modified_date)
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

class CollectionList:
    '''A list of collections

    :param count: the total count of all collections in the database
    :type count:int
    :param collections: the collections matching the criteria
    :type collections:List[Collection]
    '''
    
    def __init__(self, count: int, collections: List[Collection]) -> None:
        if not isinstance(count, int):
            raise TypeError("count must be an int")

        if not isinstance(collections, list):
            raise TypeError("collections must be a list")

        for collection in collections:
            if not isinstance(collection, Collection):
                raise TypeError("At least one item in collections is not an object of type Collection")

        self._count: int = count
        self._collections: List[Collection] = collections

    @property
    def count(self):
        '''The total number of collections in the database that match the criteria'''
        return self._count

    @property
    def collections(self):
        '''The (paged) collections matching the criteria'''
        return self._collections

    @classmethod
    def from_dict(cls, obj:dict):
        """Get a collection list initialized from a dict

        :param obj: the dictionary representing this object
        :type obj:dict

        :return: a CollectionList object
        :rtype: CollectionList
        """
        collections = [Collection.from_dict(item) for item in obj["collections"]]
        return CollectionList(obj["count"], collections)

    def to_dict(self) -> dict:
        '''
        Return a dictionary representation of this object

        :return: a dictionary representing this object
        :rtype: dict
        '''
        return {
            "count": self.count,
            "collections": [item.to_dict() for item in self.collections],
        }

    def __repr__(self) -> str:
        return str(self.to_dict())


